#include "stdafx.h"
#include "funcs.h"

// szFileName:          the name of the *.mca file
// lpLm4Items:          pointer to an array of items
// lpulItems:           in: number of items according to the size of the passed array, out: number of evaluated items
// lpunTimeUnitLength:  out: time unit length in nanoseconds read from the file
// bAllEvents:          If TRUE, all events are considered. If FALSE, only the events that causes counts within the spectrum are considered.

unsigned short ReadAndConvert(LPSTR szFileName,LPLM4_ITEM lpLm4Items,unsigned long* lpulItems,unsigned short* lpunTimeUnitLength,BOOL bAllEvents)
{ CFile infile;

  if(infile.Open(szFileName,CFile::modeRead))
  { unsigned char auchBuffer[20];

    BOOL bSpectrum;
    unsigned long ulBytesRemaining,ulTimeDifference;
    unsigned short unHeaderSize;
    unsigned char uchAdcPipelineLatency;
    unsigned short unEvent;

    unsigned long ulItemsMax=*lpulItems;

    *lpulItems=0L;

    infile.Read(auchBuffer,14);

    if(memcmp(auchBuffer,"MCA527BIN_APP ",14))  // File identication
    { // Wrong file identication.
    
      infile.Close();
    
      return RAC_NO_LM4;
    }

    infile.Read(&unHeaderSize,2);

    infile.Seek(26,CFile::begin);

    infile.Read(auchBuffer,2); 

    if(*(unsigned short*)auchBuffer!=6) // General mode must be 6.
    { // Wrong general mode.
    
      infile.Close();
    
      return RAC_NO_LM4;
    }

    infile.Seek(60,CFile::begin);

    infile.Read(lpunTimeUnitLength,2); // time unit length in nano seconds
    
    infile.Seek(220,CFile::begin);

    infile.Read(&uchAdcPipelineLatency,1); // ADC pipeline latency

    infile.Read(auchBuffer,2);

    infile.Seek(unHeaderSize,CFile::begin);

    ulBytesRemaining=infile.GetLength()-(unsigned long)unHeaderSize;

    __int64 n64TimeSinceTrigger=(__int64)-uchAdcPipelineLatency; // Because of the ADC pipeline latency, the list starts some samples before the trigger signal.

    switch(*(unsigned short*)auchBuffer) // Time coding method
    { case 0:

        while(ulBytesRemaining>=2)
        { infile.Read(auchBuffer,2);

          ulBytesRemaining-=2;
          
          if(auchBuffer[0]==0xC0)
          { // long time without event
          
            n64TimeSinceTrigger+=67907776;

            // The second byte within auchBuffer[1] is not needed and is discarded for now. It is reread in the next loop.  

            infile.Seek(infile.GetPosition()-1L,CFile::begin);

            ulBytesRemaining+=1;

            continue;
          }
          else if(auchBuffer[0]>0x88 || (auchBuffer[0] & 0x40))
          { // Unexpected value. The file is inconsistent.
          
            infile.Close();

            return RAC_INCONSISTENT;
          }
          else if(auchBuffer[0] & 0x80)
          { // The event does not cause a count within the spectrum
          
            // big-endian --> little-endian
          
            unEvent=0;
            *((unsigned char*)&unEvent+1)=auchBuffer[0];

            bSpectrum=FALSE;

            // auchBuffer[1] already contains the first byte of the time information,
          }
          else
          { // The event causes a count within the spectrum.
          
            // big-endian --> little-endian
          
            *(unsigned char*)&unEvent=auchBuffer[1];
            *((unsigned char*)&unEvent+1)=auchBuffer[0];

            bSpectrum=TRUE;

            if(ulBytesRemaining<1)
            { // Too less bytes remaining. The file is inconsistent.
            
              infile.Close();

              return RAC_INCONSISTENT;
            }

            // Read the first byte of the time information to auchBuffer[1]. 

            infile.Read(auchBuffer+1,1);

            ulBytesRemaining--;
          }
        
          // Evaluate the time information

          if((auchBuffer[1] & 0xFC)==0xFC)
          { // 4 bytes time information

            if(ulBytesRemaining<3)
            { // Too less bytes remaining. The file is inconsistent.
            
              infile.Close();

              return RAC_INCONSISTENT;
            }

            ulTimeDifference=798912L+((unsigned long)(auchBuffer[1] & 0x03)<<24);

            infile.Read(auchBuffer,3);

            ulBytesRemaining-=3;

            ulTimeDifference+=(((unsigned long)auchBuffer[0]<<16)+((unsigned long)auchBuffer[1]<<8)+(unsigned long)auchBuffer[2]);
          }
          else if((auchBuffer[1] & 0xF0)==0xF0)
          { // 3 bytes time information

            if(ulBytesRemaining<2)
            { // Too less bytes remaining. The file is inconsistent.
            
              infile.Close();

              return RAC_INCONSISTENT;
            }

            ulTimeDifference=12480L+((unsigned long)(auchBuffer[1] & 0x0F)<<16);

            infile.Read(auchBuffer,2);

            ulBytesRemaining-=2;

            ulTimeDifference+=(((unsigned long)auchBuffer[0]<<8)+(unsigned long)auchBuffer[1]);
          }
          else if((auchBuffer[1] & 0xC0)==0xC0)
          { // 2 bytes time information

            if(ulBytesRemaining<1)
            { // Too less bytes remaining. The file is inconsistent.
            
              infile.Close();

              return RAC_INCONSISTENT;
            }

            ulTimeDifference=192L+((unsigned long)(auchBuffer[1] & 0x3F)<<8);

            infile.Read(auchBuffer,1);

            ulBytesRemaining--;

            ulTimeDifference+=(unsigned long)auchBuffer[0];
          }
          else
          { // 1 byte time information

            ulTimeDifference=(unsigned long)auchBuffer[1];
          }

          n64TimeSinceTrigger+=(__int64)ulTimeDifference;

          if(bAllEvents || bSpectrum)
          { lpLm4Items[*lpulItems].unEvent=unEvent;
          
            lpLm4Items[(*lpulItems)++].nTime=n64TimeSinceTrigger;

            if(*lpulItems>=ulItemsMax && ulBytesRemaining>=2)
            { // The provided memory for items is used up.
            
              infile.Close();

              return RAC_MEM_USED_UP;
            }
          }
        }

        if(ulBytesRemaining==1)
        { infile.Read(auchBuffer,1);

          if(auchBuffer[0]!=0xC0)
          { // A last remeaining byte can only be equal to 0xC0 otherwise the file is inconsistent.
          
            infile.Close();

            return RAC_INCONSISTENT;
          }
        }

        infile.Close();

        return RAC_COMPLETE;;

      case 1:

        while(ulBytesRemaining)
        { infile.Read(auchBuffer,1);

          ulBytesRemaining--;
          
          if((auchBuffer[0] & 0xC0)==0xC0)
          { // long time without event

            n64TimeSinceTrigger+=(__int64)(256L*((unsigned long)(auchBuffer[0] & 0x3F)+1));

            continue;
          }
          else if(auchBuffer[0]>0x88 || (auchBuffer[0] & 0x40))
          { // Unexpected value. The file is inconsistent.
          
            infile.Close();

            return RAC_INCONSISTENT;
          }
          else if(auchBuffer[0] & 0x80)
          { // The event does not cause a count within the spectrum
          
            if(ulBytesRemaining<1)
            { // Too less bytes remaining. The file is inconsistent.
            
              infile.Close();

              return RAC_INCONSISTENT;
            }
            // big-endian --> little-endian
          
            unEvent=0;
            *((unsigned char*)&unEvent+1)=auchBuffer[0];

            bSpectrum=FALSE;
          }
          else
          { // The event causes a count within the spectrum.

            if(ulBytesRemaining<2)
            { // Too less bytes remaining. The file is inconsistent.
            
              infile.Close();

              return RAC_INCONSISTENT;
            }

            infile.Read(auchBuffer+1,1);

            ulBytesRemaining--;
          
            // big-endian --> little-endian
          
            *(unsigned char*)&unEvent=auchBuffer[1];
            *((unsigned char*)&unEvent+1)=auchBuffer[0];

            bSpectrum=TRUE;
          }
        
          // Evaluate the time information

          infile.Read(auchBuffer,1);

          ulBytesRemaining--;

          n64TimeSinceTrigger+=(__int64)auchBuffer[0];

          if(bAllEvents || bSpectrum)
          { lpLm4Items[*lpulItems].unEvent=unEvent;
          
            lpLm4Items[(*lpulItems)++].nTime=n64TimeSinceTrigger;

            if(*lpulItems>=ulItemsMax)
            { infile.Close();

              return RAC_MEM_USED_UP;
            }
          }
        }

        infile.Close();

        return RAC_COMPLETE;

      case 2:

        while(ulBytesRemaining)
        { infile.Read(auchBuffer,1);

          ulBytesRemaining--;
          
          if((auchBuffer[0] & 0xC0)==0xC0)
          { // long time without event

            n64TimeSinceTrigger+=(__int64)(65536L*((unsigned long)(auchBuffer[0] & 0x3F)+1));

            continue;
          }
          else if(auchBuffer[0]>0x88 || (auchBuffer[0] & 0x40))
          { // Unexpected value. The file is inconsistent.
          
            infile.Close();

            return RAC_INCONSISTENT;
          }
          else if(auchBuffer[0] & 0x80)
          { // The event does not cause a count within the spectrum
          
            if(ulBytesRemaining<2)
            { // Too less bytes remaining. The file is inconsistent.
            
              infile.Close();

              return RAC_INCONSISTENT;
            }
            // big-endian --> little-endian
          
            unEvent=0;
            *((unsigned char*)&unEvent+1)=auchBuffer[0];

            bSpectrum=FALSE;
          }
          else
          { // The event causes a count within the spectrum.

            if(ulBytesRemaining<3)
            { // Too less bytes remaining. The file is inconsistent.
            
              infile.Close();

              return RAC_INCONSISTENT;
            }

            infile.Read(auchBuffer+1,1);

            ulBytesRemaining--;
          
            // big-endian --> little-endian
          
            *(unsigned char*)&unEvent=auchBuffer[1];
            *((unsigned char*)&unEvent+1)=auchBuffer[0];

            bSpectrum=TRUE;
          }
        
          // Evaluate the time information

          infile.Read(auchBuffer,2);

          ulBytesRemaining-=2;

          n64TimeSinceTrigger+=(__int64)*(unsigned short*)auchBuffer;

          if(bAllEvents || bSpectrum)
          { lpLm4Items[*lpulItems].unEvent=unEvent;
          
            lpLm4Items[(*lpulItems)++].nTime=n64TimeSinceTrigger;

            if(*lpulItems>=ulItemsMax)
            { infile.Close();

              return RAC_MEM_USED_UP;
            }
          }
        }

        infile.Close();

        return RAC_COMPLETE;

      default:

        infile.Close();
    
        return RAC_INCONSISTENT; 
        
    }
  }  

  return RAC_INCONSISTENT;
}